local AddonName, a = ...
if a.BuildFail(40000, 50000) then return end
local L = a.Localize
local s = SpellFlashAddon
local c = BittensSpellFlashLibrary

local GetComboPoints = GetComboPoints
local GetPowerRegen = GetPowerRegen
local GetTime = GetTime
local SPELL_POWER_ECLIPSE = SPELL_POWER_ECLIPSE
local math = math
local select = select

a.Rotations = {}
c.RegisterForEvents(a)
s.Spam[AddonName] = function()
	c.Flash(a)
end

----------------------------------------------------------------------- Balance
local function calcPower(
	goingUp, startPower, startDelay, endDelay, countNextCast)
	
	if goingUp then
		return startPower
			+ 20 * c.CountLandings(
				"Starfire", startDelay, endDelay, countNextCast)
			+ 15 * c.CountLandings("Starsurge", startDelay, endDelay)
	else
		return startPower
			- 14 * c.CountLandings("Wrath", startDelay, endDelay, countNextCast)
			- 15 * c.CountLandings("Starsurge", startDelay, endDelay)
	end
end

a.Rotations.Balance = {
	Spec = 1,
	OffSwitch = "balance_off",
	
	FlashInCombat = function()
		
		-- calculate state for when the next cast will start
		local power = s.Power("player", SPELL_POWER_ECLIPSE)
		local goingUp = c.HasBuff("Eclipse (Lunar)")
			 or (not c.HasBuff("Eclipse (Solar)") and power >= 0)
		local busyTime = c.GetBusyTime()
		power = calcPower(goingUp, power, -3, busyTime)
		local eclipse
		if goingUp then
			goingUp = power < 100
			eclipse = power < 0 or power >= 100
		else
			goingUp = power <= -100
			eclipse = power > 0 or power <= -100
		end
		local eclipseSoon
		local futurePower = calcPower(
			goingUp, power, busyTime, 3 + busyTime, true)
		if goingUp then
			eclipseSoon = not eclipse and futurePower >= 100
		else
			eclipseSoon = not eclipse and futurePower <= -100
		end
		
		--"Force of Nature"
		--"Starfall if Lunar"
		--"Lunar Moonfire with Glyph of Starfire, unless eclipse w/in 2 casts"
		--"Insect Swarm, unless eclipse w/in 2 seconds"
		--"Moonfire, unless eclipse w/in 2 seconds"
		--"Starsurge unless eclipse w/in 2 seconds"
		--"Starfire if Going Up"
		--"Wrath"
		
		-- flash
		c.FlashAll("Force of Nature")
		if eclipse and c.HasBuff("Eclipse (Lunar)") then
			c.FlashAll("Starfall")
		end
		if eclipseSoon then
			if c.GetMyDebuffDuration("Insect Swarm") < 3
				or (c.GetMyDebuffDuration("Moonfire") < 3
					and c.GetMyDebuffDuration("Sunfire") < 3) then
				
				if c.FlashAll("Starsurge") then
					return
				end
			end
		else
			if eclipse and not goingUp and c.HasTalent("Sunfire") then
				if c.PriorityFlash("Insect Swarm", "Sunfire", "Starsurge") then
					return
				end
			else
				if c.PriorityFlash("Insect Swarm", "Moonfire", "Starsurge") then
					return
				end
			end
		end
		if goingUp then
			if calcPower(
				goingUp, 
				power, 
				busyTime, 
				busyTime + c.GetCastTime("Wrath")) >= 100 then
				
				c.FlashAll("Wrath")
			else
				c.FlashAll("Starfire")
			end
		else
			if calcPower(
				goingUp, 
				power, 
				busyTime, 
				busyTime + c.GetCastTime("Starfire")) <= -100 then
				
				c.FlashAll("Starfire")
			else
				c.FlashAll("Wrath")
			end
		end
	end,
	
	FlashAlways = function()
		c.FlashAll("Moonkin Form", "Mark of the Wild")
	end,
	
	Energized = function(spellID, amount)
        if math.abs(amount) > 20 and math.abs(amount) < 41 then
            c.Debug("Event", "Euphoria from", spellID, "gives", amount)
        end
	end,
	
	SpellMissed = function(spellID)
        if spellID == c.GetID("Wrath")
            or spellID == c.GetID("Starsurge")
            or spellID == c.GetID("Starfire") then
            
            c.Debug("Event", "Miss", spellID)
        end
	end
}

------------------------------------------------------------------------- feral
a.Rotations.FeralCaster = {
	Spec = 2,
	
	CheckFirst = function()
		return (not s.Form())
			and (not a.GetConfig("bear_off") or not a.GetConfig("cat_off"))
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end
}

-------------------------------------------------------------------------- bear
local uncontrolledMitigationBuffs = c.GetIDs(
	"Fury of the Beast LFR",
	"Fury of the Beast",
	"Fury of the Beast Heroic",
	"Master Tactician LFR",
	"Master Tactician",
	"Master Tactician Heroic")

a.Rotations.Bear = {
	Spec = 2,
	OffSwitch = "bear_off",
	Form = "Bear Form",
	
	FlashInCombat = function()
		
		-- Not on the GCD
		c.FlashAll("Enrage", "Growl", "Challenging Roar")
		if c.WearingSet(2, "FeralT13") then
			c.RotateCooldowns(
				uncontrolledMitigationBuffs, 
				"Berserk for Mitigation",
				"Barkskin", 
				"Survival Instincts", 
				"Frenzied Regeneration")
		else
			c.FlashAll("Berserk")
			c.RotateCooldowns(
				uncontrolledMitigationBuffs, 
				"Barkskin", 
				"Survival Instincts", 
				"Frenzied Regeneration")
		end
		local rage = s.PowerPercent("player")
		if rage > 70 then
			c.FlashAll("Maul")
		end
		
		-- Rotation
		c.PriorityFlash(
			"Mangle(Bear Form) for Debuff",
			"Demoralizing Roar",
			"Pulverize",
			"Mangle(Bear Form)",
			"Lacerate to 3",
			"Faerie Fire (Feral) to 3",
			"Thrash",
			"Faerie Fire (Feral)",
			"Lacerate")
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end,
}

--------------------------------------------------------------------------- cat
local lastRipTime = 0
local lastRipDuration = 0
local pendingRipExtension = 0

local function updateRipDuration()
	if c.IsAuraPendingFor("Rip") then
		if c.HasGlyph("Bloodletting") then
			a.RipDuration = 22
		else
			a.RipDuration = 16
		end
		c.Debug("Bloodletting", "Rip pending:", a.RipDuration)
	end
	
	if not c.HasGlyph("Bloodletting") then
		a.RipDuration = c.GetMyDebuffDuration("Rip")
		return
	end
	
	local now = GetTime()
	local duration = s.MyDebuffDuration(c.GetID("Rip"))
--	if now - lastRipTime > 1 then
--		c.Debug("Glyph of Bloodletting", "New Application")
--		pendingRipExtension = 6
--	else
	if duration - lastRipDuration > 4 then
		c.Debug("Bloodletting", "Overwrite: ", duration - lastRipDuration)
		pendingRipExtension = 6
	elseif duration > lastRipDuration then
		pendingRipExtension = pendingRipExtension - 2
		c.Debug(
			"Bloodletting",
			"change of ", duration - lastRipDuration,
			" -> ", pendingRipExtension)
	end
	lastRipTime = now
	lastRipDuration = duration
	
	duration = c.GetMyDebuffDuration("Rip")
	if duration > 0 then
		a.RipDuration = duration + pendingRipExtension
	else
		a.RipDuration = 0
	end
end

local function updateResources()
	a.Regen = select(2, GetPowerRegen())
	a.Energy = c.GetPower(a.Regen)
	
	local info = c.GetQueuedInfo()
	a.CP = GetComboPoints("player")
	if c.InfoMatches(
		info, "Shred", "Rake", "Mangle(Cat Form)", "Ravage", "Ravage!") then
		
		a.CP = math.min(5, a.CP + 1)
c.Debug("Resources", "Combo Point Pending:", a.CP)
	elseif c.InfoMatches(info, "Rip", "Savage Roar", "Ferocious Bite") then
		a.CP = 0
c.Debug("Resources", "Finisher Pending: 0")
	end
end

a.Rotations.Cat = {
	Spec = 2,
	OffSwitch = "cat_off",
	Form = "Cat Form",
	
	FlashInCombat = function()
		updateRipDuration()
		updateResources()
		
		-- Not on the GCD
		c.FlashAll(
			"Survival Instincts under 30", 
			"Barkskin under 30",
			"Tiger's Fury",
			"Berserk as Cat")
		
		-- Rotation
--c.Debug("Flash",
		c.PriorityFlash(
			"Faerie Fire (Feral) to 3 - Optional",
			"Mangle(Cat Form) for Bleed Debuff",
			"Ferocious Bite to save Rip",
			"Rip",
			"Savage Roar",
			"Ferocious Bite during execute",
			"Ferocious Bite conservative",
			"Rake",
			"Ravage",
			"Ravage!",
			"Feral Charge(Cat Form)",
			"Shred")
--, a.CP, a.Energy)
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end,
	
	AuraApplied = function(spellID)
		if spellID == c.GetID("Stampede") then
			a.SuppressRavage = false
			c.Debug("Event", "Allow Ravage")
		end
	end,
	
	CastSucceeded = function(info)
		if c.InfoMatches(info, "Ravage!") then
			a.SuppressRavage = true
			c.Debug("Event", "Deny Ravage")
		end
	end,
	
	SpellMissed = function(spellID)
		if spellID == c.GetID("Ravage!") then
			a.SuppressRavage = false
			c.Debug("Event", "Re-allow Ravage")
		end
	end,
}

------------------------------------------------------------------------- resto
local lastNG = 0
a.Rotations.Restoration = {
	Spec = 3,
	OffSwitch = "resto_off",
	
	FlashInCombat = function()
		local ngcool = GetTime() - lastNG
		if ngcool > 45 and c.HasBuff("Nature's Grace") then
			lastNG = GetTime()
		elseif ngcool >= 60 then
			local regrowth = c.GetSpell("Regrowth")
			if c.GetBuffDuration("Clearcasting")
				> c.GetCastTime("Regrowth") then
				
				regrowth.FlashColor = nil
				regrowth.FlashSize = nil
			else
				regrowth.FlashColor = "yellow"
				regrowth.FlashSize = s.FlashSizePercent() / 2
			end
			c.FlashAll("Regrowth")
		end
		
		c.FlashAll("Innervate")
	end,
	
	FlashAlways = function()
		c.FlashAll("Mark of the Wild")
	end,
}
