local AddonName, a = ...
if a.BuildFail(40000, 50000) then return end
local L = a.Localize
local s = SpellFlashAddon
local c = BittensSpellFlashLibrary

local select = select
local math = math

c.Init(a)

------------------------------------------------------------------------ Common
c.AddOptionalSpell("Mark of the Wild", nil, {
	NoRangeCheck = 1,
	CheckFirst = function()
		return c.RaidBuffNeeded(c.FIVE_PERCENT_BUFFS)
	end
})

----------------------------------------------------------------------- Balance
c.AddOptionalSpell("Moonkin Form", nil, {
	Type = "form",
})

c.AddSpell("Insect Swarm", nil, {
	EarlyRefresh = 0, -- needs to be initialized for CheckFirst
	CheckFirst = function(z)
		return c.GetMyDebuffDuration("Insect Swarm") <= z.EarlyRefresh
	end
})
c.ManageDotRefresh("Insect Swarm", 2)

local function moonfireCheckFirst(moonfireEarlyRefresh, sunfireEarlyRefresh)
	return c.GetMyDebuffDuration("Moonfire") <= moonfireEarlyRefresh
		and c.GetMyDebuffDuration("Sunfire") <= sunfireEarlyRefresh
		and (c.GetMyDebuffDuration("Insect Swarm")
				> 3 + c.GetSpell("Insect Swarm").EarlyRefresh
			or c.IsCastingOrInAir("Insect Swarm"))
end

c.AddSpell("Moonfire", nil, {
	EarlyRefresh = 0, -- needs to be initialized for CheckFirst
	FlashID = { "Moonfire", "Sunfire" },
	CheckFirst = function(z)
		return moonfireCheckFirst(z.EarlyRefresh, 0)
	end
})
c.ManageDotRefresh("Moonfire", 2)

c.AddSpell("Sunfire", nil, {
	ID = "Moonfire", -- for some reason, using moonfire's id works better
	EarlyRefresh = 0, -- needs to be initialized for CheckFirst
	FlashID = { "Moonfire", "Sunfire" }, -- Moonfire/Sunfire
	CheckFirst = function(z)
		return moonfireCheckFirst(0, z.EarlyRefresh)
	end
})
c.ManageDotRefresh("Sunfire", 2, c.GetID("Sunfire"))

c.AddSpell("Starsurge", nil, {
	NotIfActive = 1,
})

c.AddOptionalSpell("Force of Nature", nil, {
	NoRangeCheck = 1,
})

c.AddOptionalSpell("Starfall")

c.AssociateTravelTimes(1.3, "Wrath", "Starsurge")
c.AssociateTravelTimes(.1, "Insect Swarm", "Moonfire", "Sunfire", "Starfire")

------------------------------------------------------------------------- Feral
c.AddOptionalSpell("Berserk")

c.AddSpell("Faerie Fire (Feral)", "to 3", {
	Debuff = c.ARMOR_DEBUFFS,
	Stack = 3,
	EarlyRefresh = 30,
})

-------------------------------------------------------------------------- Bear
c.AddOptionalSpell("Survival Instincts", nil, {
	NoGCD = true,
})

c.AddOptionalSpell("Barkskin", nil, {
	NoGCD = true,
})

c.AddOptionalSpell("Berserk", "for Mitigation", {
	NoGCD = true,
	Buff = "Berserk",
	CheckFirst = function()
		return c.GetBuffDuration("Pulverize", false) > 8
			and c.GetMyDebuffDuration("Lacerate", false) > 9
			and s.MyDebuffStack("Lacerate") == 3
			and not c.IsAuraPendingFor("Pulverize")
	end
})

c.AddOptionalSpell("Enrage", nil, {
	NoGCD = true,
})

c.AddOptionalSpell("Frenzied Regeneration", nil, {
	NoGCD = true,
	CheckFirst = function()
		return s.HealthPercent("player") < 90
	end
})

c.AddOptionalSpell("Maul", nil, {
	NoGCD = true,
})

c.AddSpell("Demoralizing Roar", nil, {
	Debuff = c.PHYSICAL_DAMAGE_DEBUFFS,
	EarlyRefresh = 3,
	Melee = true,
})

c.AddSpell("Thrash", nil, {
	Melee = true, 
	NoRangeCheck = true,
})

c.AddSpell("Lacerate", "to 3", {
	CheckFirst = function()
		return c.GetMyDebuffStack("Lacerate") < 3
			or c.IsAuraPendingFor("Pulverize")
	end
})

c.AddSpell("Pulverize", nil, {
	CheckFirst = function()
		return c.GetMyDebuffStack("Lacerate") == 3
			and c.GetBuffDuration("Pulverize") < 1.5
	end
})

c.AddSpell("Mangle(Bear Form)", "for Debuff", {
	Debuff = c.SLOW_MELEE_DEBUFFS,
	CheckFirst = function()
		return c.HasTalent("Infected Wounds")
	end
})

c.AddSpell("Growl", nil, {
	NoGCD = true,
	CheckFirst = c.CheckFirstForTaunts,
})

c.AddSpell("Challenging Roar", nil, {
	Melee = true,
	CheckFirst = c.CheckFirstForTaunts,
})

--------------------------------------------------------------------------- Cat
local function capWarning()
	return a.Energy + a.Regen > 100
end

c.AddSpell("Survival Instincts", "under 30", {
	NoGCD = true,
	FlashColor = "red",
	CheckFirst = function()
		return s.HealthPercent("player") < 30
	end
})

c.AddSpell("Barkskin", "under 30", {
	NoGCD = true,
	FlashColor = "red",
	CheckFirst = function()
		return s.HealthPercent("red") < 30
	end
})

c.AddOptionalSpell("Feral Charge(Cat Form)", nil, {
	NoRangeCheck = 1,
	CheckFirst = function()
		return c.HasTalent("Stampede") and not c.HasBuff("Berserk")
	end
})

c.AddOptionalSpell("Tiger's Fury", nil, {
	NoGCD = true,
	CheckFirst = function()
		local bonus = 20 * c.GetTalentRank("King of the Jungle")
		return (bonus == 0 or a.Energy < 90 - bonus) 
			and a.CP < 5
			and not c.HasBuff("Stampede")
	end
})

c.AddSpell("Berserk", "as Cat", {
	NoGCD = true,
	CheckFirst = function()
		return a.Energy >= 80 and c.GetCooldown("Tiger's Fury") > 3
	end
})

c.AddOptionalSpell("Faerie Fire (Feral)", "to 3 - Optional", {
	Debuff = c.ARMOR_DEBUFFS,
	Stack = 3,
	EarlyRefresh = 30,
})

c.AddSpell("Mangle(Cat Form)", "for Bleed Debuff", {
	CheckFirst = function(z)
		local duration = 
			math.max(0, s.DebuffDuration(c.BLEED_DEBUFFS) - c.GetBusyTime())
		local rake = c.GetMyDebuffDuration("Rake")
		if math.abs(duration - rake) < 1 then
			return duration < 3
		else
			return duration < 2
		end
	end
})

c.AddSpell("Rip", nil, {
	CheckFirst = function()
		return a.CP == 5 and c.GetMyDebuffDuration("Rip") < 1.9
	end
})

c.AddSpell("Savage Roar", nil, {
	NoRangeCheck = 1,
	CheckFirst = function()
		if c.GetBuffDuration("Savage Roar") > a.RipDuration + 3 then
			return false
		end
		
		local buildup = a.CatSimulator.TimeToFive(25, "Savage Roar")
		return buildup < a.RipDuration
			and (buildup + 1.5 > a.RipDuration
				or (a.CP == 5
					and (not c.HasBuff("Savage Roar")
						or capWarning()
						or c.HasBuff("Berserk"))))
	end
})

c.AddSpell("Ferocious Bite to save Rip", nil, {
	ID = 22568,
	CheckFirst = function()
		local maxHealth
		if c.WearingSet(2, "FeralT13") then
			maxHealth = 60
		else
			maxHealth = 25
		end
		return a.RipDuration > .2 
			and a.RipDuration < 3
			and s.HealthPercent() < maxHealth
			and c.HasTalent("Blood in the Water")
	end
})

c.AddSpell("Ferocious Bite", "conservative", {
	CheckFirst = function()
		if a.CP < 5 then
			return false
		end
		
		if c.GetBuffDuration("Savage Roar") < a.RipDuration + 3 then
			return false
		end
		
		return a.RipDuration > a.CatSimulator.TimeToFive(35, "FB conservative")
	end
})

c.AddSpell("Ferocious Bite", "during execute", {
	CheckFirst = function()
		if c.WearingSet(2, "FeralT13") then
			maxHealth = 60
		else
			maxHealth = 25
		end
		return a.RipDuration > .2 
			and a.CP == 5
			and s.HealthPercent() < maxHealth
			and c.HasTalent("Blood in the Water")
	end
})

c.AddSpell("Rake", nil, {
	CheckFirst = function()
		return c.GetMyDebuffDuration("Rake") < 1.9
			and not c.IsAuraPendingFor("Rake")
	end
})

local function ravageOverride(z)
	duration = c.GetBuffDuration("Stampede")
	local points = a.CP
	return not a.SuppressRavage
		and duration > 0
		and s.MeleeDistance() 
		and (duration < 3 or (points < 5 and a.Energy < 80))
		and not s.SpellDelay(z.FlashID)
end

c.AddSpell("Ravage", nil, {
	FlashID = {"Ravage", "Ravage!"},
	Override = ravageOverride,
})

c.AddSpell("Ravage!", nil, {
	FlashID = {"Ravage", "Ravage!"},
	Override = ravageOverride,
})

local function order(t1, t2)
	if t1 < t2 then
		return t1, t2
	else
		return t2, t1
	end
end

c.AddSpell("Shred", nil, {
	CheckFirst = function()
		local firstRefresh, secondRefresh = order(
			c.GetMyDebuffDuration("Rake"),
			s.DebuffDuration(c.BLEED_DEBUFFS) - c.GetBusyTime())
		local finisherRefresh = c.GetBuffDuration("Savage Roar")
		if a.RipDuration < finisherRefresh + 3 then
			finisherRefresh = a.RipDuration
		end
		
		-- shred to avoid capping
		if capWarning() 
			and (finisherRefresh == 0 or finisherRefresh > 2.5)
			and (firstRefresh > 2.5 or a.CP < 4)
			and (secondRefresh > 3.5 or a.CP < 3) then
			
			return true
		end
		
		-- pool if we will have enough combo points without shred
		if not c.HasBuff("Berserk")
			and (a.CP == 5
				or (a.CP == 4 and firstRefresh < finisherRefresh)
				or (a.CP == 3 and secondRefresh < finisherRefresh)) then
				
			return false
		end
		
		-- shred unless it would not leave enough energy to refresh things
		a.CatSimulator.Init("shred")
		a.CatSimulator.PoolThenCast(40)
		a.CatSimulator.PoolThenCast(35)
		if a.CatSimulator.Time > firstRefresh + 1 then
c.Debug("Shred", "Fail first refresh", a.CatSimulator.Time, firstRefresh)
			return false
		end
		a.CatSimulator.PoolThenCast(35)
		if a.CatSimulator.Time > secondRefresh + 1 then
c.Debug("Shred", "Fail second refresh", a.CatSimulator.Time, secondRefresh)
			return false
		end
		return true
	end
})

------------------------------------------------------------------------- Resto
c.AddOptionalSpell("Innervate", nil, {
	NoRangeCheck = true,
	CheckFirst = function()
		return s.PowerPercent("player") < 75 and c.HasBuff("Power Torrent")
	end
})

c.AddOptionalSpell("Regrowth", nil, {
	NoRangeCheck = true,
})
